"""Subpackage to provide a baseline for implementing materials in geometries.

Usage of this module's contents is optional for geometry development with
*legend-pygeom-tools*, it is completely equivalent to manual construction and
registration of materials.
"""

from __future__ import annotations

from abc import ABC
from functools import wraps
from typing import Callable

import pyg4ometry.geant4 as g4


def cached_property(material: Callable):
    @wraps(material)
    def wrapper(self):
        attr = f"_{material.__name__}"
        if not hasattr(self, attr):
            setattr(self, attr, material(self))
        return getattr(self, attr)

    return property(wrapper)


class BaseMaterialRegistry(ABC):  # noqa: B024
    def __init__(self, g4_registry: g4.Registry):
        self.g4_registry = g4_registry

        self._elements = {}
        self._elements_cb = {}
        self._define_elements()

    def get_element(self, symbol: str) -> g4.Element:
        if (symbol in self._elements_cb) and (symbol not in self._elements):
            self._elements[symbol] = (self._elements_cb[symbol])()
        return self._elements[symbol]

    def _add_element(self, name: str, symbol: str, z: int, a: float) -> None:
        """Lazily define an element on the current registry."""
        assert symbol not in self._elements_cb
        self._elements_cb[symbol] = lambda: g4.ElementSimple(
            name=name, symbol=symbol, Z=z, A=a, registry=self.g4_registry
        )

    def _define_elements(self) -> None:
        """Lazily define all used elements."""
        self._add_element(name="Hydrogen", symbol="H", z=1, a=1.00794)
        self._add_element(name="Boron", symbol="B", z=5, a=10.811)
        self._add_element(name="Carbon", symbol="C", z=6, a=12.011)
        self._add_element(name="Nitrogen", symbol="N", z=7, a=14.01)
        self._add_element(name="Oxygen", symbol="O", z=8, a=16.00)
        self._add_element(name="Fluorine", symbol="F", z=9, a=19.00)
        self._add_element(name="Sodium", symbol="Na", z=11, a=22.99)
        self._add_element(name="Aluminium", symbol="Al", z=13, a=26.981539)
        self._add_element(name="Silicon", symbol="Si", z=14, a=28.09)
        self._add_element(name="argon", symbol="Ar", z=18, a=39.95)
        self._add_element(name="Chromium", symbol="Cr", z=24, a=51.9961)
        self._add_element(name="Manganese", symbol="Mn", z=25, a=54.93805)
        self._add_element(name="Iron", symbol="Fe", z=26, a=55.845)
        self._add_element(name="Indium", symbol="In", z=49, a=114.82)
        self._add_element(name="Cobalt", symbol="Co", z=27, a=58.9332)
        self._add_element(name="Nickel", symbol="Ni", z=28, a=58.6934)
        self._add_element(name="Copper", symbol="Cu", z=29, a=63.55)
        self._add_element(name="Tantalum", symbol="Ta", z=73, a=180.94)
        self._add_element(name="Gold", symbol="Au", z=79, a=196.967)



"""Assignment of sensitive detectors to physical volumes, for use in ``remage``."""

from __future__ import annotations

import json
import logging
from collections.abc import Generator
from dataclasses import dataclass
from itertools import groupby
from pathlib import Path
from typing import Literal

import pyg4ometry.geant4 as g4
from dbetto import AttrsDict
from pyg4ometry.gdml.Defines import Auxiliary

log = logging.getLogger(__name__)

AUXKEY_DETMETA = "RMG_detector_meta"
AUXKEY_DET = "RMG_detector"


@dataclass
class RemageDetectorInfo:
    detector_type: Literal["optical", "germanium", "scintillator"]
    """``remage`` detector type."""

    uid: int
    """``remage`` detector UID."""

    metadata: object | None = None
    """Attach arbitrary metadata to this sensitive volume. This will be written to GDML as JSON.

    See also
    ========
    .get_sensvol_metadata
    """


def walk_detectors(
    pv: g4.PhysicalVolume | g4.LogicalVolume | g4.Registry,
) -> Generator[tuple[g4.PhysicalVolume, RemageDetectorInfo], None, None]:
    """Iterate over all physical volumes that have a :class:`RemageDetectorInfo` attached.

    Important
    ---------
    this only returns instances previously set via
    :meth:`get_pygeom_active_detector() <pyg4ometry.geant4.PhysicalVolume.get_pygeom_active_detector>`,
    not data loaded from GDML. Use :meth:`get_all_sensvols` instead for that use case.
    """

    if isinstance(pv, g4.PhysicalVolume):
        det = pv.get_pygeom_active_detector()
        if det is not None:
            assert isinstance(det, RemageDetectorInfo)
            yield pv, det
    if isinstance(pv, g4.LogicalVolume):
        next_v = pv
    if isinstance(pv, g4.PhysicalVolume):
        next_v = pv.logicalVolume
    elif isinstance(pv, g4.Registry):
        next_v = pv.worldVolume
    else:
        msg = "invalid type encountered in walk_detectors volume tree"
        raise TypeError(msg)

    for dv in next_v.daughterVolumes:
        if dv.type == "placement":
            yield from walk_detectors(dv)


def generate_detector_macro(registry: g4.Registry, filename: str) -> None:
    """Create a Geant4 macro file containing the defined active detector volumes for use in remage."""
    if _get_rmg_detector_aux(registry, raise_on_missing=False) is not None:
        sensvols = get_all_sensvols(registry)
    else:
        sensvols = {pv.name: det for pv, det in walk_detectors(registry)}

    macro_lines = {}
    for pv, det in sensvols.items():
        if pv in macro_lines:
            continue
        mac = f"/RMG/Geometry/RegisterDetector {det.detector_type.title()} {pv} {det.uid}\n"
        macro_lines[pv] = mac

    macro_contents = "".join(macro_lines.values())

    with Path(filename).open("w", encoding="utf-8") as f:
        f.write(macro_contents)


def write_detector_auxvals(registry: g4.Registry) -> None:
    """Append an auxiliary structure, storing the sensitive detector volume information.

    .. note::
        see :doc:`../metadata` for a reference of the written structure.
    """
    if _get_rmg_detector_aux(registry, raise_on_missing=False) is not None:
        msg = "detector auxiliary structure already written"
        raise RuntimeError(msg)

    written_pvs = set()
    group_it = groupby(
        sorted(walk_detectors(registry), key=lambda d: d[1].detector_type),
        lambda d: d[1].detector_type,
    )

    meta_group_aux = Auxiliary(AUXKEY_DETMETA, "", registry)

    for key, group in group_it:
        group_aux = Auxiliary(AUXKEY_DET, key, registry)

        for pv, det in group:
            if pv.name in written_pvs:
                continue
            written_pvs.add(pv.name)

            group_aux.addSubAuxiliary(
                Auxiliary(pv.name, det.uid, registry, addRegistry=False)
            )
            if det.metadata is not None:
                json_meta = json.dumps(det.metadata)
                meta_group_aux.addSubAuxiliary(
                    Auxiliary(pv.name, json_meta, registry, addRegistry=False)
                )


def _get_rmg_detector_aux(
    registry: g4.Registry, *, raise_on_missing: bool = True
) -> Auxiliary | None:
    auxs = [aux for aux in registry.userInfo if aux.auxtype == AUXKEY_DETMETA]
    if auxs == []:
        if not raise_on_missing:
            return None
        msg = f"GDML missing {AUXKEY_DETMETA} auxval (not written by legend-pygeom-tools?)"
        raise RuntimeError(msg)
    assert len(auxs) == 1
    return auxs[0]


def get_sensvol_metadata(registry: g4.Registry, name: str) -> AttrsDict | None:
    """Load metadata attached to the given sensitive volume (from GDML)."""
    meta_aux = _get_rmg_detector_aux(registry)
    meta_auxs = [aux for aux in meta_aux.subaux if aux.auxtype == name]
    if meta_auxs == []:
        return None
    assert len(meta_auxs) == 1
    return AttrsDict(json.loads(meta_auxs[0].auxvalue))


def get_all_sensvols(registry: g4.Registry) -> dict[str, RemageDetectorInfo]:
    """Load all registered sensitive detectors with their metadata (from GDML)."""
    meta_aux = _get_rmg_detector_aux(registry)
    meta_auxs = {
        aux.auxtype: AttrsDict(json.loads(aux.auxvalue)) for aux in meta_aux.subaux
    }

    detmapping = {}
    type_auxs = [aux for aux in registry.userInfo if aux.auxtype == AUXKEY_DET]
    for type_aux in type_auxs:
        for det_aux in type_aux.subaux:
            detmapping[det_aux.auxtype] = RemageDetectorInfo(
                type_aux.auxvalue, int(det_aux.auxvalue), meta_auxs.get(det_aux.auxtype)
            )

    if set(meta_auxs.keys()) - set(detmapping.keys()) != set():
        msg = "invalid GDML auxval structure (meta keys and detmapping keys differ)"
        raise RuntimeError(msg)

    return detmapping


def get_sensvol_by_uid(
    registry: g4.Registry, uid: int
) -> tuple[str, RemageDetectorInfo] | None:
    """Get the volume name and detector metadata for the detector with remage detector ID `uid`."""
    sensvols = get_all_sensvols(registry)
    found = next(filter(lambda s: s[1].uid == uid, sensvols.items()), None)
    if found is None:
        return None
    return found


def __set_pygeom_active_detector(self, det_info: RemageDetectorInfo | None) -> None:
    """Set the remage detector info on this physical volume instance."""
    if not isinstance(self, g4.PhysicalVolume):
        msg = "patched-in function called on wrong type"
        raise TypeError(msg)
    assert self.registry is not None
    if _get_rmg_detector_aux(self.registry, raise_on_missing=False) is not None:
        msg = "detector auxiliary structure already written"
        raise RuntimeError(msg)
    self.__pygeom_active_detector = det_info


def __get_pygeom_active_detector(self) -> RemageDetectorInfo | None:
    """Get the remage detector info on this physical volume instance."""
    if not isinstance(self, g4.PhysicalVolume):
        msg = "patched-in function called on wrong type"
        raise TypeError(msg)

    if hasattr(self, "__pygeom_active_detector"):
        return self.__pygeom_active_detector
    return None


def __patch_pyg4_pv():
    """monkey-patch a new function onto every PhysicalVolume instance."""
    t = g4.PhysicalVolume
    t.set_pygeom_active_detector = __set_pygeom_active_detector
    t.get_pygeom_active_detector = __get_pygeom_active_detector

    prop = property(__get_pygeom_active_detector)
    t.pygeom_active_detector = prop.setter(__set_pygeom_active_detector)

    def _wrap_warn(fn):
        def _fn(*args):
            import warnings

            warnings.warn(
                "pygeom_active_dector (typo!) is deprecated, use pygeom_active_detector instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return fn(*args)

        return _fn

    prop = property(_wrap_warn(__get_pygeom_active_detector))
    t.pygeom_active_dector = prop.setter(_wrap_warn(__set_pygeom_active_detector))


__patch_pyg4_pv()



from __future__ import annotations

import logging
from pathlib import Path

import pyg4ometry.geant4 as g4
from pyg4ometry.gdml.Defines import Auxiliary

log = logging.getLogger(__name__)


def _color_macro_recursive(lv: g4.LogicalVolume, macro_lines: dict) -> None:
    if hasattr(lv, "pygeom_color_rgba") and lv.name not in macro_lines:
        mac = f"/vis/geometry/set/forceSolid {lv.name}\n"
        if lv.pygeom_color_rgba is False or lv.pygeom_color_rgba[3] == 0:
            mac += f"/vis/geometry/set/visibility {lv.name} -1 false\n"
        else:
            rgba = " ".join([str(c) for c in lv.pygeom_color_rgba])
            mac += f"/vis/geometry/set/colour {lv.name} 0 {rgba}\n"
        macro_lines[lv.name] = mac

    for pv in lv.daughterVolumes:
        if pv.type == "placement":
            _color_macro_recursive(pv.logicalVolume, macro_lines)


def generate_color_macro(registry: g4.Registry, filename: str) -> None:
    """Create a Geant4 macro file containing the defined visualization attributes."""
    macro_lines = {registry.worldVolume: None}
    _color_macro_recursive(registry.worldVolume, macro_lines)
    macro_contents = "".join([m for m in macro_lines.values() if m is not None])

    with Path(filename).open("w", encoding="utf-8") as f:
        f.write(macro_contents)


def write_color_auxvals(registry: g4.Registry) -> None:
    """Append an auxiliary structure, storing the visualization color information."""
    written_lvs = set()

    def _append_color_recursive(lv: g4.LogicalVolume) -> None:
        if hasattr(lv, "pygeom_color_rgba") and lv.name not in written_lvs:
            if lv.pygeom_color_rgba is False or lv.pygeom_color_rgba[3] == 0:
                rgba = "-1"
            else:
                rgba = ",".join([str(c) for c in lv.pygeom_color_rgba])
            # remove existing colors.
            lv.auxiliary = [aux for aux in lv.auxiliary if aux.auxtype != "rmg_color"]
            lv.addAuxiliaryInfo(
                Auxiliary("rmg_color", rgba, registry, addRegistry=False)
            )
            written_lvs.add(lv.name)

        if hasattr(lv, "pygeom_colour_rgba"):
            msg = f"pygeom_colour_rgba on volume {lv.name} not supported, use use pygeom_color_rgba instead."
            raise RuntimeError(msg)

        for pv in lv.daughterVolumes:
            if pv.type == "placement":
                _append_color_recursive(pv.logicalVolume)

    written_lvs.add(registry.worldVolume.name)  # do not store world vis args.
    _append_color_recursive(registry.worldVolume)


def load_color_auxvals_recursive(lv: g4.LogicalVolume) -> None:
    auxvals = list(filter(lambda aux: aux.auxtype == "rmg_color", lv.auxiliary))
    if len(auxvals) > 1:
        log.warning("more than one rmg_color for LV %s", lv.name)
    # assert len(auxvals) <= 1

    if len(auxvals) > 0 and not hasattr(lv, "pygeom_color_rgba"):
        rgba = auxvals[-1].auxvalue
        if rgba == "-1":
            lv.pygeom_color_rgba = False
        else:
            lv.pygeom_color_rgba = list(map(float, rgba.split(",")))

    for pv in lv.daughterVolumes:
        if pv.type == "placement":
            load_color_auxvals_recursive(pv.logicalVolume)



# file generated by setuptools-scm
# don't change, don't track in version control

__all__ = ["__version__", "__version_tuple__", "version", "version_tuple"]

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Tuple
    from typing import Union

    VERSION_TUPLE = Tuple[Union[int, str], ...]
else:
    VERSION_TUPLE = object

version: str
__version__: str
__version_tuple__: VERSION_TUPLE
version_tuple: VERSION_TUPLE

__version__ = version = '0.0.15'
__version_tuple__ = version_tuple = (0, 0, 15)



from __future__ import annotations

import os

from pyg4ometry import gdml, geant4

from . import detectors, geometry, visualization


def write_pygeom(
    reg: geant4.Registry,
    gdml_file: str | os.PathLike | None = None,
    write_vis_auxvals: bool = True,
) -> None:
    """Commit all auxiliary data to the registry and write out a GDML file."""
    detectors.write_detector_auxvals(reg)
    if write_vis_auxvals:
        visualization.write_color_auxvals(reg)
    geometry.check_registry_sanity(reg, reg)

    if gdml_file is not None:
        # pyg4ometry has added color writing in their bdsim style by default in 2025.
        try:
            w = gdml.Writer(writeColour=False)
        except TypeError:
            w = gdml.Writer()

        w.addDetector(reg)
        w.write(str(gdml_file))



"""Helper functions for geometry construction."""

from __future__ import annotations

from pyg4ometry import geant4


def check_registry_sanity(v, registry: geant4.Registry) -> None:
    """Check recursively if all children in the volume and material tree have the correct
    registry instance attached.

    Parameters
    ==========
    v
        object to recursively check to have the right registry.
    registry
        the expected registry to compare against

    Note
    ====
    This function prevents an easy-to-miss problem using pyg4ometry: If different (or no)
    registries are used inside an object structure, this might lead to unexpected results
    in GDML output.
    """
    if not isinstance(v, geant4.Registry) and v.registry is not registry:
        msg = f"found invalid registry instance on {v}"
        raise RuntimeError(msg)

    # walk the tree.
    if isinstance(v, (geant4.LogicalVolume, geant4.AssemblyVolume)):
        for dv in v.daughterVolumes:
            check_registry_sanity(dv, registry)
        check_registry_sanity(v.material, registry)
        check_registry_sanity(v.solid, registry)

    elif isinstance(v, geant4.PhysicalVolume):
        check_registry_sanity(v.logicalVolume, registry)

    elif isinstance(v, geant4.Registry):
        check_registry_sanity(v.worldVolume, registry)
        for _, s in v.surfaceDict.items():
            check_registry_sanity(s, registry)

    elif isinstance(v, geant4.Material):
        if hasattr(v, "components"):
            for comp in v.components:
                check_registry_sanity(comp[0], registry)

    elif isinstance(v, geant4.SurfaceBase):
        check_registry_sanity(v.surface_property, registry)

    elif isinstance(
        v, (geant4.solid.OpticalSurface, geant4.solid.SolidBase, geant4.Element)
    ):
        pass

    else:
        msg = f"invalid type {type(v)} encountered in check_registry_sanity volume tree"
        raise TypeError(msg)
